(******************************************************************************)
(* plx parser                                                                 *)
(* (C) 2010 Steffen Solyga <solyga@absinth.net>                               *)
(******************************************************************************)
IMPLEMENTATION MODULE Prsr;

IMPORT FIO;
FROM Scnr IMPORT Put, Read, KeepId, Diff, TSymbol, sym, num, id, pos;
FROM Comp IMPORT TValue, TOperation, Val, Opr, Sto, Lod;

CONST
  err ::= FIO.ErrorOutput;
  cntMax = 32;

TYPE
  TClass = ( cConst, cVar, cType );
  TType = ( tBool, tInt );
  TObjDesc = RECORD
    id : CARDINAL;
    typ: TType;
    CASE cls: TClass OF
    | cConst: val: TValue;
    | cVar  : idx: CARDINAL; ok: BOOLEAN;
    | cType :
    END;
  END;

VAR
  objs: ARRAY [ 0 .. cntMax - 1 ] OF TObjDesc;
  cnt: CARDINAL;
  idx: CARDINAL;
  oprs: ARRAY TSymbol OF TOperation;

PROCEDURE Err( s: ARRAY OF CHAR );
BEGIN
  FIO.WrStr( err, 'Prsr (' );
  FIO.WrCard( err, pos.l, 0 );
  FIO.WrChar( err, ',' );
  FIO.WrCard( err, pos.c, 0 );
  FIO.WrStr( err, '): ' );
  FIO.WrStr( err, s );
  FIO.WrStr( err, '.' );
  FIO.WrLn( err );
  HALT;
END Err;

PROCEDURE Find( VAR i: CARDINAL );
BEGIN
  i := 0;
  WHILE i < cnt DO
    IF Diff( objs[i].id, id ) = 0 THEN RETURN END;
    INC( i );
  END;
END Find;


PROCEDURE Expression( VAR t: TType; VAR v: TValue; c: BOOLEAN );

  PROCEDURE SimpleExpr( VAR t: TType; VAR v: TValue; c: BOOLEAN );

    PROCEDURE Factor( VAR t: TType; VAR v: TValue; c: BOOLEAN );
    VAR i: CARDINAL;
    BEGIN
      CASE sym OF
      | sNum : IF num > MAX( TValue ) THEN Err( 'Number too large' ) END;
               t := tInt;
               IF c THEN v := num ELSE Val( TValue(num) ) END;
               Read();
      | sId  : Find( i );
               IF i >= cnt THEN Err( 'Identifier undeclared' ) END;
               t := objs[ i ].typ;
               IF c THEN
                 IF objs[ i ].cls # cConst THEN Err( 'Expected constant' ) END;
                 v := objs[ i ].val;
               ELSE
                 CASE objs[ i ].cls OF
                 | cConst: Val( objs[i].val );
                 | cVar  : IF ~ objs[i].ok THEN
                             Err( 'Variable not initialised' )
                           END;
                           Lod( objs[i].idx );
                 ELSE
                   Err( 'Expected constant or variable' );
                 END;
               END;
               Read();
      | sLPar: Read(); Expression( t, v, c );
               IF sym # sRPar THEN Err( 'Expected operator or ")"' ) END;
               Read();
      | sNot : Read(); Factor( t, v, c );
               IF t # tBool THEN Err( 'Expected boolean factor' ) END;
               IF c THEN
                 IF v = 0 THEN v := 1 ELSE v := 0 END;
               ELSE
                 Opr( oprs[sNot] );
               END;
      ELSE
        Err( 'Expected factor' );
      END;
    END Factor;

    PROCEDURE Term( VAR t: TType; VAR v: TValue; c: BOOLEAN );
    VAR s: TSymbol; t1: TType; v1: TValue;
    BEGIN
      Factor( t, v, c );
      WHILE ( sMul <= sym ) & ( sym <= sAnd ) DO
        IF ( sym = sAnd ) # ( t = tBool ) THEN Err( 'Wrong operator type' ) END;
        s := sym; Read(); Factor( t1, v1, c );
        IF t # t1 THEN Err( 'Wrong argument type' ) END;
        IF c THEN
          CASE s OF
          | sMul, sAnd: v := v * v1;
          | sDiv      : v := v DIV v1;
          ELSE
            Err( 'Bug: Invalid mulop' );
          END;
        ELSE
          Opr( oprs[s] );
        END;
      END;
    END Term;

  VAR s: TSymbol; t1: TType; v1: TValue;
  BEGIN
    IF ( sym = sPlus ) OR ( sym = sMinus ) THEN
      s := sym; Read(); Term( t, v, c );
      IF t # tInt THEN Err( 'Expected integer term' ) END;
      IF s = sMinus THEN
        IF c THEN v := -v ELSE Opr( oNeg ) END;
      END;
    ELSE
      Term( t, v, c );
    END;
    WHILE ( sPlus <= sym ) & ( sym <= sOr ) DO
      IF ( sym = sOr ) # ( t = tBool ) THEN Err( 'Wrong operator type' ) END;
      s := sym; Read(); Term( t1, v1, c );
      IF t # t1 THEN Err( 'Wrong argument type' ) END;
      IF c THEN
        CASE s OF
        | sPlus : v := v + v1;
        | sMinus: v := v - v1;
        | sOr   : IF ( v # 0 ) OR ( v1 # 0 ) THEN v := 1 ELSE v := 0 END;
        ELSE
          Err( 'Bug: Invalid addop' );
        END;
      ELSE
        Opr( oprs[s] );
      END;
    END;
  END SimpleExpr;

VAR s: TSymbol; t1: TType; v1: TValue;
BEGIN
  SimpleExpr( t, v, c );
  IF ( sEql <= sym ) & ( sym <= sGeq ) THEN
    s := sym; Read(); SimpleExpr( t1, v1, c );
    IF t # t1 THEN Err( 'Wrong argument type' ) END;
    IF c THEN
      CASE s OF
      | sEql: IF v = v1  THEN v := 1 ELSE v := 0 END;
      | sNeq: IF v # v1  THEN v := 1 ELSE v := 0 END;
      | sLss: IF v < v1  THEN v := 1 ELSE v := 0 END;
      | sLeq: IF v <= v1 THEN v := 1 ELSE v := 0 END;
      | sGtr: IF v > v1  THEN v := 1 ELSE v := 0 END;
      | sGeq: IF v >= v1 THEN v := 1 ELSE v := 0 END;
      ELSE
        Err( 'Bug: Invalid relop' );
      END;
    ELSE
      Opr( oprs[s] );
    END;
    t := tBool;
  END;
END Expression;

PROCEDURE Statement();
VAR i: CARDINAL; t: TType; v: TValue;
BEGIN
  CASE sym OF
  | sId : (* assignment *)
          Find( i );
          IF i >= cnt THEN Err( 'Identifier undeclared' ) END;
          IF objs[ i ].cls # cVar THEN Err( 'Not a variable' ) END;
          Read();
          IF sym # sBec THEN Err( 'Expected ":="' ) END;
          Read(); Expression( t, v, FALSE );
          IF t # objs[i].typ THEN Err( 'Incompatible types' ) END;
          Sto( objs[i].idx );
          objs[ i ].ok := TRUE;
  | sWrt: (* expression output *)
          Read(); Expression( t, v, FALSE );
          CASE t OF
          | tInt : Opr( oWrt );
          | tBool: Opr( oWrB );
          ELSE
            Err( 'Bug: Invalid type' );
          END;
  ELSE
    (* empty *)
  END;
END Statement;

PROCEDURE ConstDeclaration();
VAR i: CARDINAL;
BEGIN  (* sym = sId *)
  Find( i );
  IF i < cnt THEN Err( 'Identifier redeclared' ) END;
  IF cnt >= cntMax THEN Err( 'Too many identifiers' ) END;
  objs[ i ].id  := id;
  objs[ i ].cls := cConst;
  KeepId();
  Read();
  IF sym # sEql THEN Err( 'Expected "="' ) END;
  Read();
  Expression( objs[i].typ, objs[i].val, TRUE );
  INC( cnt );
  IF sym # sSemi THEN Err( 'Expected ";"' ) END;
  Read();
END ConstDeclaration;

PROCEDURE VarDeclaration();
VAR i,j: CARDINAL;
BEGIN  (* sym = sId *)
  Find( i );
  IF i < cnt THEN Err( 'Identifier redeclared' ) END;
  IF cnt >= cntMax THEN Err( 'Too many identifiers' ) END;
  objs[ i ].id  := id;
  objs[ i ].cls := cVar;
  KeepId();
  Read();
  IF sym # sCol THEN Err( 'Expected ":"' ) END;
  Read();
  IF sym # sId THEN Err( 'Expected identifier' ) END;
  Find( j );
  IF j >= cnt THEN Err( 'Identifier undeclared' ) END;
  IF objs[ j ].cls # cType THEN Err( 'Not a type' ) END;
  objs[ i ].typ := objs[ j ].typ;
  objs[ i ].idx := idx;
  objs[ i ].ok  := FALSE;
  INC( cnt ); INC( idx );
  Read();
  IF sym # sSemi THEN Err( 'Expected ";"' ) END;
  Read();
END VarDeclaration;

PROCEDURE Run(); (* Program() *)
BEGIN
  Read();
  IF sym = sConst THEN
    Read();
    WHILE sym = sId DO ConstDeclaration() END;
  END;
  IF sym = sVar THEN
    Read();
    WHILE sym = sId DO VarDeclaration() END;
  END;
  IF sym # sBeg THEN Err( 'Expected "BEGIN"' ) END;
  Read(); Statement();
  WHILE sym = sSemi DO Read(); Statement() END;
  IF sym # sDot THEN Err( 'Expected ";" or "."' ) END;
(*  Read(); IF sym # sEOF THEN Err( 'Expected EOF' ) END; *)
END Run;

PROCEDURE InitOprs();
VAR s: TSymbol;
BEGIN
  FOR s := MIN( TSymbol ) TO MAX( TSymbol ) DO oprs[ s ] := oNull END;
(*  oprs[ sExcl  ] := oWrt; *)
  oprs[ sWrt   ] := oWrt;
  oprs[ sNot   ] := oNot;
  oprs[ sMul   ] := oMul;
  oprs[ sDiv   ] := oDiv;
  oprs[ sAnd   ] := oAnd;
  oprs[ sPlus  ] := oAdd;
  oprs[ sMinus ] := oSub;
  oprs[ sOr    ] := oOr;
  oprs[ sEql   ] := oEql;
  oprs[ sNeq   ] := oNeq;
  oprs[ sLss   ] := oLss;
  oprs[ sLeq   ] := oLeq;
  oprs[ sGtr   ] := oGtr;
  oprs[ sGeq   ] := oGeq;
END InitOprs;

BEGIN
  cnt := 0;
  idx := 0;
  Put( 'FALSE' );
  objs[ cnt ].id  := id;
  objs[ cnt ].typ := tBool;
  objs[ cnt ].cls := cConst;
  objs[ cnt ].val := 0;
  KeepId(); INC( cnt );
  Put( 'TRUE' );
  objs[ cnt ].id  := id;
  objs[ cnt ].typ := tBool;
  objs[ cnt ].cls := cConst;
  objs[ cnt ].val := 1;
  KeepId(); INC( cnt );
  Put( 'BOOLEAN' );
  objs[ cnt ].id  := id;
  objs[ cnt ].typ := tBool;
  objs[ cnt ].cls := cType;
  KeepId(); INC( cnt );
  Put( 'INTEGER' );
  objs[ cnt ].id  := id;
  objs[ cnt ].typ := tInt;
  objs[ cnt ].cls := cType;
  KeepId(); INC( cnt );
  InitOprs();
END Prsr.
